"""Daemon for alise."""

# vim: tw=100 foldmethod=indent
# pylint: disable = logging-fstring-interpolation

import sys
import os
import uvicorn

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles

from fastapi_oauth2.router import router as oauth2_router

from alise.logsetup import logger

# from alise.config import CONFIG
# from alise.parse_args import args

from alise.pathdiversion import SSROAuth2Middleware

from alise.oauth2_config import oauth2_config
from alise.router_api import router_api
from alise.router_ssr import router_ssr

# from alise.marcus_oauth2 import router as marcus_oauth2_router


app = FastAPI()

app.include_router(router_ssr)
app.include_router(router_api)
app.include_router(oauth2_router)

static_path = os.path.join(os.path.dirname(__file__), "static")
app.mount("/static", StaticFiles(directory=static_path), name="static")

# for r in router_api.routes:
#     logger.debug(f"router_api.routes: {r}")
# logger.debug(f"router_api.routes: {type(router_api.routes)}")
# for r in router_ssr.routes:
#     logger.debug(f"router_ssr.routes: {r}")
# logger.debug(f"router_ssr.routes: {type(router_ssr.routes)}")

# app.add_middleware(OAuth2Middleware, config=oauth2_config, callback=on_auth)
# app.add_middleware(OAuth2Middleware, config=oauth2_config)


# app.add_middleware(SSROAuth2Middleware, config=oauth2_config, callback=on_auth)
app.add_middleware(SSROAuth2Middleware, config=oauth2_config)


def main():
    """Console script for alise."""
    # logger.debug("This is just a test for 'debug'")
    # logger.info("This is just a test for 'info'")
    # logger.warning("This is just a test for 'warning'")
    # logger.error("This is just a test for 'error'")

    # uvicorn.run(root, host="0.0.0.0", port=4711)
    # uvicorn.run(root, host="0.0.0.0", port=8000, log_level="info")
    uvicorn.run(app, host="0.0.0.0", port=8000, log_level="debug")
    logger.debug("--------startup done----------")


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# vim: tw=100 foldmethod=indent

import logging
from logging.handlers import RotatingFileHandler
import sys

from alise.parse_args import args
from alise.config import CONFIG

# logger = logging.getLogger(__name__)
logger = logging.getLogger("")  # => This is the key to allow logging from other modules

DISPLAY_PATHLEN = 15


class PathTruncatingFormatter(logging.Formatter):
    """formatter for logging"""

    def format(self, record):
        pathname = record.pathname
        if len(pathname) > DISPLAY_PATHLEN:
            pathname = f"...{pathname[-(DISPLAY_PATHLEN+2):]}"
        record.pathname = pathname
        return super().format(record)


def setup_logging():
    """setup logging"""

    formatter = logging.Formatter(
        fmt="[%(asctime)s.%(msecs)03d]%(levelname)8s - %(message)s", datefmt="%H:%M:%S"
    )

    formatter = PathTruncatingFormatter(
        fmt=f"[%(asctime)s] [%(pathname){DISPLAY_PATHLEN}s:%(lineno)-4d]%(levelname)8s - %(message)s",
        datefmt="%H:%M:%S",
    )

    # setup logfile
    try:
        logfile = args.logfile
    except AttributeError:
        logfile = None
    if logfile is None:
        logfile = CONFIG.messages.log_file
    if logfile:
        handler = RotatingFileHandler(logfile, maxBytes=10**6, backupCount=2)
    else:
        handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # setup log level
    try:
        loglevel = args.loglevel
    except AttributeError:
        loglevel = None
    if loglevel is None:
        loglevel = CONFIG.messages.log_level
    if loglevel is None:
        loglevel = "INFO"
    logger.setLevel(loglevel)

    # turn off logging noise:
    my_level = logging.CRITICAL
    logging.getLogger().setLevel(logging.DEBUG)
    logging.getLogger("flask_pyoidc").setLevel(my_level)
    logging.getLogger("oic").setLevel(my_level)
    logging.getLogger("jwkest").setLevel(my_level)
    logging.getLogger("urllib3").setLevel(my_level)
    logging.getLogger("werkzeug").setLevel(my_level)
    logging.getLogger("flaat").setLevel(my_level)
    logging.getLogger("httpx").setLevel(my_level)
    return logger


logger = setup_logging()



# vim: tw=100 foldmethod=indent
# pylint: disable = logging-fstring-interpolation

import os
from sys import exit
import requests
from dotenv import load_dotenv
import hashlib
import json

# from social_core.backends.github import GithubOAuth2
from social_core.backends.google import GoogleOAuth2

# from social_core.backends.elixir import ElixirOpenIdConnect
from social_core.backends.open_id_connect import OpenIdConnectAuth

from fastapi_oauth2.claims import Claims
from fastapi_oauth2.config import OAuth2Config
from fastapi_oauth2.client import OAuth2Client

from alise.config import CONFIG
from alise.logsetup import logger
from alise import exceptions

from ssl import SSLCertVerificationError


CONFIG_KEY_MAP = {
    "ACCESS_TOKEN_URL": "token_endopint",
    "AUTHORIZATION_URL": "authorization_endpoint",
    "REVOKE_TOKEN_URL": "revocation_endpoint",
    "USERINFO_URL": "userinfo_endpoint",
    "JWKS_URI": "jwks_uri",
    # apparently not used: = rsp.json()["introspection_endpoint"]
}


def make_oidc_config_class_google(op_name):
    class NewClass(GoogleOAuth2):
        # needed by alise
        OIDC_ENDPOINT = "https://accounts.google.com/"
        provider_type = "external"

    return NewClass


def make_oidc_config_class(op_name, op_config):
    class NewClass(OpenIdConnectAuth):
        provider_type: str
        def setting(self, op_name, default=None):
            return getattr(self, op_name, default)

    try:
        ## Sonderlocke for google
        if op_config.op_url.startswith("https://accounts.google.com"):
            return make_oidc_config_class_google(op_name)

        NewClass.__name__ = op_name
        NewClass.name = op_name
        NewClass.name = op_name
        if op_config.op_url is None:
            logger.error(f"trying to configure op without issuer url: {op_name}")
            raise exceptions.InternalException(f"trying to configure op without issuer url: {op_name}")
        NewClass.OIDC_ENDPOINT = op_config.op_url
        NewClass.ID_TOKEN_ISSUER = NewClass.OIDC_ENDPOINT
        NewClass.provider_type = "external"
        if op_config.internal:
            NewClass.provider_type = "internal"
    except AttributeError as e:
        logger.warning(f"Cannot find attribute for {op_name}: {e}")

    ## Autoconf
    autoconf_url = op_config.op_url + "/.well-known/openid-configuration"
    if op_config.op_config_url is not "":
        logger.debug(F"setting manual config URL for {op_name}: {op_config.op_config_url}")
        autoconf_url = op_config.op_config_url
    try:
        autoconf = requests.get(autoconf_url , timeout=15).json()
    except (SSLCertVerificationError, requests.exceptions.SSLError) as e:
        if op_config.ignore_ssl_errors is True:
            logger.warning(f"Ignoring existing SSL Errors with {op_name}")
            autoconf = requests.get(autoconf_url , timeout=15, verify=False).json()
        else:
            logger.error(f"SSL Error with {op_name}: {e}")
            raise e
    try:
        NewClass.ACCESS_TOKEN_URL = autoconf["token_endpoint"]
        NewClass.AUTHORIZATION_URL = autoconf["authorization_endpoint"]
        NewClass.REVOKE_TOKEN_URL = autoconf["revocation_endpoint"]
        NewClass.USERINFO_URL = autoconf["userinfo_endpoint"]
        NewClass.JWKS_URI = autoconf["jwks_uri"]
    except KeyError as e:
        logger.warning(f"Cannot find {e} for {op_name}")
    return NewClass


def make_oidc_config(op_name):
    # def logged1_lambda(user):
    #     logger.debug(F"{user.sub=}")
    #     logger.debug(F"{user.provider}:{user.sub}")
    #     return F"{user.provider}:{user.sub}"
    def generate_username(user):
        logger.debug(F"generating username")
        logger.debug(F"claim: {op_config.username_claim=}")
        for key,value in user.items():
            logger.debug(F"   user.{key} = {value}")
        logger.debug(F"{user.get(op_config.username_claim)=}")
        return user.get(op_config.username_claim)
    op_config = CONFIG.auth.get_op_config(op_name)
    backend = make_oidc_config_class(op_name, op_config)
    logger.debug(f"going to generate_username:")
    logger.debug(f"{op_name=} - {op_config.username_claim=}")
    # logger.debug(F"found: {user.get(op_config.username_claim)}")
    client = OAuth2Client(
        backend=backend,
        client_id=op_config.client_id,
        client_secret=op_config.client_secret,
        scope=op_config.scopes,
        claims=Claims(
            identity=lambda user: f"{user.provider}:{user.sub}",
            generated_username=lambda user: generate_username(user),
        ),
    )
    return client
            # identity=lambda user: logged1_lambda(user),
            # generated_username=lambda user: generate_username(user),
            # identity=lambda user: f"{user.provider}:{user.sub}",
            # generated_username=lambda user: f"{user.get(op_config.username_claim)}",


# vega:                generated_username=lambda user: f"{user.upn}",
# fels:                generated_username=lambda user: f"{user.sub}",


configured_clients = []
for op_name in CONFIG.auth.get_op_names():
    configured_clients.append(make_oidc_config(op_name))

oauth2_config = OAuth2Config(
    allow_http=True,
    jwt_secret="secret",
    jwt_expires=900,
    jwt_algorithm="HS256",
    clients=configured_clients,
)


def get_provider_iss_by_name(name: str) -> str:
    for x in oauth2_config.clients:
        if x.backend.name == name:
            try:
                return x.backend.OIDC_ENDPOINT  # pyright: ignore
            except AttributeError:
                return ""
    return ""


def get_provider_name_by_iss(iss: str) -> str:
    for x in oauth2_config.clients:
        if x.backend.OIDC_ENDPOINT == iss:  # pyright: ignore
            return x.backend.name
    return ""


def get_provider_name_by_hash(iss: str, hash_method="sha1") -> str:
    for x in oauth2_config.clients:
        for test_string in [
            x.backend.OIDC_ENDPOINT,  # pyright: ignore
            x.backend.OIDC_ENDPOINT + "\n",  # pyright: ignore
        ]:

            hash_function = getattr(hashlib, hash_method)()
            hash_function.update((test_string).encode())  # pyright: ignore
            hash = hash_function.hexdigest()
            # logger.debug(F"hash: {hash} - {iss} - {x.backend.OIDC_ENDPOINT.encode()}")
            if hash == iss:  # pyright: ignore
                return x.backend.name
    return ""


def get_sub_iss_by_identity(identity):
    provider_name, sub = identity.split(":")
    iss = get_provider_iss_by_name(provider_name)
    return (sub, iss)


def get_provider_name_sub_by_identity(identity):
    logger.debug(f"identity: {identity}")
    provider_name, sub = identity.split(":")
    return (provider_name, sub)


def get_providers(provider_type):
    names = []
    for x in oauth2_config.clients:
        try:
            if x.backend.provider_type == provider_type:  # pyright: ignore
                names.append(x.backend.name)
        except AttributeError:
            if provider_type == "external":  # external providers may not
                names.append(x.backend.name)  # explicitly define this attribute
    return names


def get_providers_long():
    providers = {"internal": [], "external": []}
    for x in oauth2_config.clients:
        try:
            provider_type = x.backend.provider_type  # pyright: ignore
        except AttributeError as e:
            logger.warning(e)
            provider_type = "external"
        provider_name = x.backend.name
        provider_iss = x.backend.OIDC_ENDPOINT  # pyright: ignore
        providers[provider_type].append({"name": provider_name, "iss": provider_iss})
    return providers


def get_internal_providers():
    returnv = get_providers("internal")
    logger.info(f"internal providers: {returnv}")
    return returnv


def get_external_providers():
    return get_providers("external")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
#  nova documentation build configuration file
#
# Refer to the Sphinx documentation for advice on configuring this file:
#
#   http://www.sphinx-doc.org/en/stable/config.html

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

sys.path.insert(0, os.path.abspath("../../"))
sys.path.insert(0, os.path.abspath("../"))
sys.path.insert(0, os.path.abspath("./"))

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.

extensions = [
    "sphinx.ext.autodoc",
    "sphinxarg.ext",
    "sphinx_autodoc_typehints",
    "sphinx_rtd_theme",
    "pbr.sphinxext",
]

autodoc_member_order = "bysource"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

todo_include_todos = True

source_parsers = {
    # '.md': 'recommonmark.parser.CommonMarkParser',
}

# The suffix of source filenames.
source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "alise"
copyright = "2023 - present, "
author = "Marcus Hardt"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
# release = version.release_string
# The short X.Y version.
# version = version.version_string

# A list of glob-style patterns that should be excluded when looking for
# source files. They are matched against the source file names relative to the
# source directory, using slashes as directory separators on all platforms.
exclude_patterns = []

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ["alise."]

# -- Options for man page output ----------------------------------------------

# If true, show URL addresses after external links.
# man_show_urls = False

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# Add any paths that contain "extra" files, such as .htaccess or
# robots.txt.
# html_extra_path = ['_extra']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%Y-%m-%d %H:%M"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = False


# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    #    'index':    ['sidebar.html', 'sourcelink.html', 'searchbox.html'],
    #    '**':       ['sidebar.html', 'localtoc.html', 'relations.html',
    #                 'sourcelink.html', 'searchbox.html'],
    "index": [
        "about.html",
        "navigation.html",
        "relations.html",
        "sourcelink.html",
        "searchbox.html",
        "sidebarfooter.html",
    ],
}

html_always_document_param_types = True


# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
# html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = "doc"
